@@ -15,7 +15,9 @@
 #
 
 from __future__ import absolute_import, division, generators, unicode_literals, print_function, nested_scopes, \
-    with_statement
+    with_statement 
+
+from builtins import bytes
 
 import datetime
 import logging as log
@@ -32,6 +34,7 @@ from cryptography import x509
 from cryptography.x509.oid import NameOID, ExtensionOID
 from cryptography.hazmat.primitives import hashes
 import ipaddress
+import sys
 
 
 MIME_JSON = "application/json"
@@ -59,7 +62,7 @@ def log_errors(data):
 
 
 class KeyType:
-    ALLOWED_SIZES = [512, 1024, 2048, 3072, 4096, 8192]
+    ALLOWED_SIZES = [2048, 3072, 4096, 8192]
     ALLOWED_CURVES = ["p256", "p384", "p521"]
     RSA = "rsa"
     ECDSA = "ec"
@@ -114,6 +117,18 @@ class ZoneConfig:
         self.key_type = key_type
 
 
+class TokenInfo:
+    def __init__(self, access_token, expires, identity, refresh_token, refresh_until,
+                 scope, token_type):
+        self.access_token = access_token
+        self.expires = expires
+        self.identity = identity
+        self.refresh_token = refresh_token
+        self.refresh_until = refresh_until
+        self.scope = scope
+        self.token_type = token_type
+
+
 class Policy:
     def __init__(self, policy_id=None, company_id=None, name=None, system_generated=None,
                  creation_date=None, subject_cn_regexes=None, subject_o_regexes=None,
@@ -167,6 +182,8 @@ class CertificateRequest:
                  san_dns=None,
                  email_addresses="",
                  ip_addresses=None,
+                 user_principal_names=None,
+                 uniform_resource_identifiers=None,
                  attributes=None,
                  key_type=None,
                  private_key=None,
@@ -187,6 +204,8 @@ class CertificateRequest:
         :param list[str] san_dns: Alternative names for SNI.
         :param list[str] email_addresses: List of email addresses
         :param list[str] ip_addresses: List of IP addresses
+        :param list[str] user_principal_names: user principal names in username@domain format
+        :param list[str] uniform_resource_identifiers: URIs
         :param attributes:
         :param KeyType key_type: Type of asymmetric cryptography algorithm. Default is RSA 2048.
         :param asymmetric.PrivateKey private_key: String with pem encoded private key or  asymmetric.PrivateKey
@@ -202,6 +221,8 @@ class CertificateRequest:
         self.san_dns = san_dns or []
         self.email_addresses = email_addresses
         self.ip_addresses = ip_addresses or []
+        self.user_principal_names = user_principal_names or []
+        self.uniform_resource_identifiers = uniform_resource_identifiers or []
         self.attributes = attributes
         self.key_password = key_password
         self.key_type = key_type
@@ -330,6 +351,43 @@ class CertificateRequest:
             for mail in self.email_addresses:
                 alt_names.append(x509.RFC822Name(mail))
 
+        if self.uniform_resource_identifiers:
+            for uri in self.uniform_resource_identifiers:
+                alt_names.append(x509.UniformResourceIdentifier(uri))
+
+        if self.user_principal_names:
+            for upn in self.user_principal_names:
+                # Python cryptography library doesn't include
+                # OID for UPN in any of the OID constants
+                # See http://www.oid-info.com/get/1.3.6.1.4.1.311.20.2.3
+                # and https://docs.microsoft.com/en-us/windows/security/identity-protection/smart-cards/smart-card-certificate-requirements-and-enumeration
+                UPNOID = x509.ObjectIdentifier('1.3.6.1.4.1.311.20.2.3')
+                # x509 library expects DER encoded string, so encode UPN into bytes
+                # with ASN1 syntax for UTF8String, which is a type/length/value encoding
+                # per https://docs.microsoft.com/en-us/windows/win32/seccertenroll/about-utf8string
+                # Construct the array of bytes (note bytes are strings in python2)
+                # by inserting the header consisting of the tag '0x0C' followed by the length of the upn
+                if (sys.version_info > (3,0)):
+                    # For python3, since we have native bytes available we'll convert the string into bytes.
+                    # However, for cases when this method is called with inputs extracted from existing certificates
+                    # & csrs, the input will already be encoded as bytes. So we'll check the input, if it is a 
+                    # string we'll convert it into a bytes object then insert our header. Otherwise, we'll just
+                    # insert the header in the passed in bytes.
+                    if (isinstance(upn,str)):
+                        bupn = bytes(upn,'utf-8')
+                    else:
+                        bupn = upn
+                    values = [12,len(upn)]
+                    header = bytes(values)
+                    data = header + bupn
+                    alt_names.append(x509.OtherName(UPNOID,data))
+                else:
+                    # For python2, since there is no native bytes method, we'll use the bytes method 
+                    # in the future module to convert the string we constructed into bytes.
+                    bupn = ''.join(chr(x) for x in [12,len(upn)])
+                    bupn = bupn + str(upn)
+                    alt_names.append(x509.OtherName(UPNOID,bytes(bupn,'utf-8')))
+
         csr_builder = csr_builder.add_extension(
             x509.SubjectAlternativeName(alt_names),
             critical=False,
@@ -406,6 +464,24 @@ class RevocationRequest:
         self.disable = disable
 
 
+CLIENT_ID = "vcert-sdk"  # type: str
+SCOPE = "certificate:manage,revoke"  # type: str
+
+
+class Authentication:
+    def __init__(self, user=None, password=None, access_token=None, refresh_token=None, api_key=None, state=None,
+                 token_expires=None, client_id=CLIENT_ID, scope=SCOPE):
+        self.user = user
+        self.password = password
+        self.access_token = access_token
+        self.refresh_token = refresh_token
+        self.token_expires = token_expires
+        self.api_key = api_key
+        self.client_id = client_id
+        self.scope = scope
+        self.state = state
+
+
 class CommonConnection:
 
     def auth(self):
@@ -465,14 +541,18 @@ class CommonConnection:
                 log_errors(r.content)
             raise VenafiConnectionError("Server status: %s\n Response: %s" %
                                         (r.status_code, r.request.url))
+
         content_type = r.headers.get("content-type")
+        # Content-type not present, return status and reason (if any)
+        if content_type is None:
+            return r.status_code, r.reason
         if content_type.startswith(MIME_TEXT):
             log.debug(r.text)
             return r.status_code, r.text
         elif content_type.startswith(MIME_HTML):
             log.debug(r.text)
             return r.status_code, r.text
-        # content-type in respons is  application/json; charset=utf-8
+        # content-type in response is  application/json; charset=utf-8
         elif content_type.startswith(MIME_JSON):
             log.debug(r.content.decode())
             return r.status_code, r.json()
